/* 
NFSTEST/MPI -     
    MPI- 
      
 NFS-   
*/

/*   4  2003  */
const char version_date_string[] = "2003/07/04";

/*
      MPI,
    MPI,  :
 (MPI_Barrier), 
    (MPI_Bcast)
       
    ( , NFS),
        

    . 

: 
mpicc -o nfstest nfstest-mpi.c

         -DDEBUG

:
mpirun -np < > nfstest <>

    - 
        nfstest-mpi.html
  doc.

   ,    
E-mail: support@parallel.ru, http://parallel.ru/testmpi
*/

/*        
   */

static char help_message[] = 
"NFSTEST/MPI - Network filesystem performance test, implemented with MPI\n\
Usage: mpirun -np <N> nfstest <parameters>\n\
Supported parameters:\n\
u (unit) - units for file size, b, K or M\n\
m (minimum) - starting file size (in units)\n\
M (Maximum) - biggest file size (in units)\n\
b (buffer) - size of memory I/O buffer (in units)\n\
s (step) - step for file size\n\
K (coeficient) - multiplier for file size\n\
R (repeat) - number of iterations to repeat the whole test\n\
f (folder) - folder to create test files\n\
n (number) - number of files written and read by each process\n\
S (Space) - quota for total space used by test files\n\
d (direct) - flag to use direct I/O (read/write instead of fread/fwrite) \n\
v (verbose) - flag to produce verbose output\n\
o (output) - output file name (instead of stdout)\n\
c (check correctness) - flag to check correctness of input buffer\n\
Example: mpirun -np 4 nfstest n3 uK m1 M1024 f/home/tmp onfstest.res c1\n\
Meaning of output:\n\
FS - size of a single file\n\
TFS - total size of all files\n\
WTT/RTT - complete write or read test time in seconds\n\
AWT/ART - average mean time to receive or send file data (for file server)\n\
LWR/LRR - local write/read rates (for 1 process)\n\
TWT/TRT - total write/read througput (for fileserver)\n";

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <mpi.h>
#include <errno.h>
#include <unistd.h>

/*   , 
      
   .  
  ,
    */
#define MAXFILES 256
#define MAXBUFSIZE (512*1024*1024)
#define MAXNAMELEN 256

/*      */
static char* outfilenames[MAXFILES];
static char*  infilenames[MAXFILES];

/*  ,  ,      */
static char nodename[32];
static char*  nodenames;
int namelen = 32;

/*   ,   fopen() */
static FILE* files[MAXFILES];

/*        */
static char infile_checkbyte[MAXFILES];

/* ,    */
static char folder[MAXNAMELEN];
static char default_test_folder[] = ".";

/*     - ,  
 ( )  */
unsigned int fsize_unit = 1024 * 1024;
static char unit_char = 'M';
static char unit_str[] = "Mb";

/* ,    MPI - 
	     */
int nprocs, myid;

/*      */
unsigned int fsize_min = 1;
unsigned int fsize_max = 1;

/*        ( s  K) 
 ,   ,   
 ,      2 */
unsigned int fsize_step = 0;
unsigned int fsize_multiplier = 2;

/*      ( S)
 ,   */
unsigned int test_space = 0;

/*    
     fsize_unit ( b) 
   ,    */
static char* buffer;
unsigned int buf_size = 0;
size_t buffer_bytes = MAXBUFSIZE;

/*        */
int pause_time = 0;

/*   ()   */
int test_repeat = 1;

/*       ( ) */
int test_correctness = 0;

/*     read/write ( d) */
int use_direct_io = 0;

/*      ( v) */
int verbose_output = 0;

/*  ,    ( n)
 ,     nprocs */
int nfiles = 0;

/* ,       ( o)
  stdout */
static char output_file_name[MAXNAMELEN];
static FILE *out = NULL;

/*    */
double test_writing(unsigned int filesize,int nfiles);
double test_reading(unsigned int filesize,int nfiles);

int main (int argc,char *argv[])
{
	MPI_Status mpi_status;
	double t0,t1,start_time,finish_time;

	int i,j,k,m,n;
	char *s;

	unsigned int fsize, repeat_counter;
	unsigned int local_fsize, total_fsize;
	double total_time,average_file_time;
	double local_rate,total_rate;
	double best_write_rate = 0,best_read_rate = 0;
	unsigned int best_read_tfs = 0,best_write_tfs = 0;

	/*  ,       */
	if(strcmp(argv[1],"--help") == 0 || 
	strcmp(argv[1],"-h") == 0 ||
	strcmp(argv[1],"-?") == 0)
	{
		fprintf(stdout,"%s",help_message);
		fflush(stdout);
		exit(0);
	}
	
	MPI_Init(&argc,&argv);                                                                       
	MPI_Comm_size(MPI_COMM_WORLD,&nprocs);                                                       
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);                                                         

	out = stdout;
	strcpy(folder,default_test_folder);

	/*      */
	if(myid == 0)
	for(i = 1; i < argc; i ++)
		switch(argv[i][0]) 
		{
			case 'u': 
				unit_char = argv[i][1];
				if(unit_char == 'M')
				{
					fsize_unit = 1024 * 1024;
					strcpy(unit_str,"Mb");
				}
				else if(unit_char == 'K')
				{
					fsize_unit = 1024;
					strcpy(unit_str,"Kb");
				}
				else if(unit_char == 'B' || unit_char == 'b')
				{
					fsize_unit = 1;
					strcpy(unit_str,"b");
				}
				break;
			
			case 'm': 
					fsize_min = atoi(argv[i]+1); 
					break;

			case 'M': 
					fsize_max = atoi(argv[i]+1); 
					break;

			case 's': 
					fsize_step = atoi(argv[i]+1); 
					fsize_multiplier = 1; 
					break;

			case 'K': 
					fsize_multiplier = atoi(argv[i]+1); 
					fsize_step = 0; 
					break;

			case 'b': 
					buf_size = atoi(argv[i]+1); 
					break;

			case 'S': 
					test_space = atoi(argv[i]+1); 
					break;

			case 'p': 
					pause_time = atoi(argv[i]+1); 
					break;

			case 'c': 
					test_correctness = atoi(argv[i]+1); 
					break;

			case 'v': 
					verbose_output = atoi(argv[i]+1); 
					break;

			case 'd':
					use_direct_io =  atoi(argv[i]+1); 
					break;

			case 'n': 
					nfiles = atoi(argv[i]+1); 
					break;

			case 'R': 
					test_repeat = atoi(argv[i]+1); 
					break;

            case 'o': 
					strcpy(output_file_name,argv[i]+1);  
					out = fopen(output_file_name,"a+");
					if(out == NULL)
						printf("WARNING: Failed to open or create %s, writing to stdout\n",output_file_name);
					else 
						printf("Writing output to %s\n",output_file_name);
               		break;

            case 'f': 
					strcpy(folder,argv[i]+1);  
					printf("Creating test files in folder: %s\n",folder);
               		break;

            case 'h': 
			case '?':
					fprintf(out,"%s",help_message);
					fflush(out);
               		break;

			default: 
					fprintf(stderr,"WARNING: unrecognized option: %s\n",argv[i]);
					break;
		}

	/*      */
 	MPI_Bcast(&fsize_unit,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&fsize_min,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&fsize_max,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&fsize_step,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&fsize_multiplier,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&buf_size,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&test_space,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&test_repeat,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(folder,MAXNAMELEN,MPI_CHAR,0,MPI_COMM_WORLD);
 	MPI_Bcast(&unit_char,1,MPI_CHAR,0,MPI_COMM_WORLD);
 	MPI_Bcast(&pause_time,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&test_correctness,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&use_direct_io,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&verbose_output,1,MPI_INT,0,MPI_COMM_WORLD);
 	MPI_Bcast(&nfiles,1,MPI_INT,0,MPI_COMM_WORLD);

	/*     
		 ,     */

	/*      ,    
		      */
	if(nfiles == 0)
		nfiles = nprocs; 

	if((myid == 0) && verbose_output)
	{
		fprintf(out,"\nFILETEST/NFS/MPI of %s is running with following parameters:\n",version_date_string);
		fprintf(out,"\tUnit is 1%s = %d bytes\n",unit_str,fsize_unit);
		fprintf(out,"\tFile size: %u%s to %u%s, buffer size: %u%s\n",fsize_min,unit_str,fsize_max,unit_str,buf_size,unit_str);
		fprintf(out,"\tNumber of processes: %d, number of files: %d, total: %d\n",nprocs,nfiles,nprocs*nfiles);
		if(test_correctness)
			fprintf(out,"\tChecking correctness of read buffer.\n");
		if(use_direct_io)
			fprintf(out,"\tUsing direct I/O calls (read/write).\n");
		if(pause_time > 0)
                  fprintf(out,"\tPausing for %d seconds after each test.\n", pause_time);
		if(test_repeat > 1)
			fprintf(out,"\tNumber of iterations for whole test: %d\n",test_repeat);
		fprintf(out,"\tTesting file system: %s\n",folder);
		fflush(out);
	}

	if(verbose_output)
	{
	/*     ,     */

		MPI_Get_processor_name(nodename,&namelen);
		if(myid == 0)
		{
			nodenames = (char*) malloc(35*nprocs);
			strcpy(nodenames,nodename);

		/*            */

			for(i = 1; i < nprocs; i ++)
			{
				strcat(nodenames," ");
				MPI_Recv(nodename,32,MPI_CHAR,i,2003,MPI_COMM_WORLD,&mpi_status);
				strcat(nodenames,nodename);	
			}
		fprintf(out,"\tNodes: %s\n",nodenames);
		}	
		else
			MPI_Send(nodename,32,MPI_CHAR,0,2003,MPI_COMM_WORLD);
		fflush(out);
	}

	/*          */
	for(i = 0; i < nfiles; i ++)
	{
		outfilenames[i] = (char*) malloc(MAXNAMELEN);	
		 infilenames[i] = (char*) malloc(MAXNAMELEN);	

		j = (myid + i + 1) % nprocs;
		k = (myid - i - 1) % nprocs;
		if(k < 0)
			k += nprocs;
		m = i / nprocs;

		sprintf(outfilenames[i],"%s/test-%02d-writes-%02d-reads.%02d.tmp",folder,myid,j,m);
		sprintf( infilenames[i],"%s/test-%02d-writes-%02d-reads.%02d.tmp",folder,k,myid,m);

#ifdef DEBUG
	fprintf(stderr,"Process %d will write %s and read %s files (step %d)\n",myid,outfilenames[i],infilenames[i],i);
#endif
		infile_checkbyte[i] = k + '0';

		/*  ,       */
		unlink(outfilenames[i]);
	} 

	start_time = MPI_Wtime();

	/*           */

	for(repeat_counter = 0; repeat_counter < test_repeat; repeat_counter ++)
	for(fsize = fsize_min; fsize <= fsize_max; fsize = fsize * fsize_multiplier + fsize_step) 
	{
		if((myid == 0) && verbose_output)
			fprintf(out,"\nStarting test iteration = %d for FS = %d%s.\n",repeat_counter,fsize,unit_str);


	/*     ,      */

		if((buf_size == 0) || (buf_size > fsize))
			buf_size = fsize;

		/*      */
	
		buffer_bytes = buf_size * fsize_unit;

		buffer = (char*) malloc(buffer_bytes);

#ifdef DEBUG
	fprintf(stdout,"DEBUG INFO. %d allocated %d-bytes buffer\n",myid,buffer_bytes);
	fflush(out);
#endif
		if(buffer == NULL)
		{
			fprintf(stderr,"malloc() error: failed to allocate %u bytes\ntry using smaller fixed buffer\n",buffer_bytes);
			exit(1);
		}

		local_fsize = fsize * nfiles;
	    total_fsize = fsize * nprocs * nfiles;

		if((test_space > 0) && (total_fsize > test_space))
			break;	

		for(i = 0; i < buffer_bytes; i ++)
			buffer[i] = myid + '0';

#ifdef DEBUG
	fprintf(stdout,"DEBUG INFO. %d is starting writing test procedure.\n",myid);
	fflush(stdout);
#endif

		/*         */
		total_time = test_writing(fsize,nfiles);

		average_file_time = total_time / (nfiles * nprocs);
		local_rate = ((double) local_fsize) / (total_time);
		total_rate = ((double) total_fsize) / (total_time);

		if(total_rate > best_write_rate)
		{
			best_write_rate = total_rate;
			best_write_tfs = total_fsize;
		}

		if(myid == 0)
			fprintf(out,"--- TFS=%u%s\tWTT=%4.1fs AWT=%4.1fs\tLWR=%3.1f%s/s TWR=%3.1f%s/s\n",
				total_fsize, unit_str, total_time, average_file_time,local_rate, unit_str, total_rate, unit_str);

		for(i = 0; i < buffer_bytes; i ++)
			buffer[i] = -1;

#ifdef DEBUG
	fprintf(stdout,"DEBUG INFO. %d is starting reading test procedure\n",myid);
	fflush(stdout);
#endif

		total_time = test_reading(fsize,nfiles);

		average_file_time = total_time / (nfiles * nprocs);
		local_rate = ((double) local_fsize) / (total_time);
		total_rate = ((double) total_fsize) / (total_time);

		if(total_rate > best_read_rate)
		{
			best_read_rate = total_rate;
			best_read_tfs = total_fsize;
		}

		if(myid == 0)
			fprintf(out,"--- TFS=%u%s\tRTT=%4.1fs ART=%4.1fs\tLRR=%3.1f%s/s TRR=%3.1f%s/s\n",
				total_fsize, unit_str, total_time,average_file_time,local_rate,unit_str, total_rate, unit_str);

		for(i = 0; i < nfiles; i ++)
			unlink(outfilenames[i]);

		free(buffer);

		fflush(out);
		fflush(stdout);

		if(pause_time)
			sleep(pause_time);
	}

	finish_time = MPI_Wtime();

	if(myid == 0)
	{
		if(verbose_output)
			fprintf(out,"\nTesting complete in %5.1f seconds\n",finish_time - start_time);
		fprintf(out,"--- Best writing rate achieved: TWR=%4.1f %s/s (at TFS=%d%s)\n",best_write_rate,unit_str,best_write_tfs,unit_str);
		fprintf(out,"--- Best reading rate achieved: TRR=%4.1f %s/s (at TFS=%d%s)\n",best_read_rate,unit_str,best_read_tfs,unit_str);
	}
	
	MPI_Finalize();
	exit(0);
	return 0;
}


double test_writing(unsigned int fsize,int nfiles)
{
	int units_left,units_to_do,units_done;
	size_t bytes_to_do,bytes_done;
	int fd;
	double t0,t1;
	int i;
	
	MPI_Barrier(MPI_COMM_WORLD);

	t0 = MPI_Wtime();

	for (i = 0; i < nfiles; i++)
	{
#ifdef DEBUG
	fprintf(stdout,"DEBUG INFO. %d is trying to open %s for writing.\n",myid,outfilenames[i]);
#endif

		files[i] = fopen(outfilenames[i], "w" );
		if(use_direct_io)
			fd = fileno(files[i]);

		if (files[i] == NULL)
		{
			fprintf(stderr,"\nfopen() failed to open %s for writing\n", outfilenames[i]);
			fflush(stderr);
			MPI_Abort(MPI_COMM_WORLD,1);
			exit(1);
		}

		units_left = fsize;

#ifdef DEBUG
	fprintf(stdout,"DEBUG INFO. %d opened file %s for writing. Using %d unit buffer\n",myid,outfilenames[i],buf_size);
#endif

		/*    */
		while(units_left > 0)
		{
			units_to_do = (units_left > buf_size) ? buf_size : units_left;

#ifdef DEBUG
	fprintf(stdout,"DEBUG INFO. %d is writing %d units to %s (total %d left)\n",myid,units_to_do,outfilenames[i],units_left);
#endif
			/*     -  (write)   (fwrite) */
			if(use_direct_io)
			{
				bytes_to_do = units_to_do * fsize_unit;
				bytes_done = write(fd,buffer,bytes_to_do);
				if(bytes_done != bytes_to_do)
				{
					fprintf(stderr,"\nERROR on %d (%s). write() = %d failed to write %d bytes to %s (fsize=%d, units_left=%d)\nERROR %d: %s\n",
                                                myid,nodename,(int)bytes_done,(int)bytes_to_do,infilenames[i],fsize,units_left,errno,strerror(errno));
					MPI_Abort(MPI_COMM_WORLD,1);
					exit(1);
				}
			}
			else
			{
             	units_done = fwrite( buffer, fsize_unit, units_to_do, files[i]);    
				if(units_done != units_to_do)
				{
					fprintf(stderr,"\nERROR on %d (%s). fwrite() = %d failed to write %u units to %s (fsize=%d, units_left=%d)\nERROR %d: %s\n",
									myid,nodename,units_done,units_to_do,outfilenames[i],fsize,units_left,errno,strerror(errno));
					fflush(stderr);
					MPI_Abort(MPI_COMM_WORLD,1);
					exit(1);
				}
			}
			units_left -= units_to_do;
		} /*   while()    */

		fclose(files[i]);
#ifdef DEBUG
	fprintf(stdout,"DEBUG INFO. %d closed file %s.\n",myid,outfilenames[i]);
#endif

	 } /*   for()    */

	MPI_Barrier(MPI_COMM_WORLD);

	t1 = MPI_Wtime();
	return t1 - t0;	
}


double test_reading(unsigned int fsize,int nfiles)
{
	int units_left,units_to_do,units_done;
	size_t bytes_to_do,bytes_done;
	int fd;
	char expected_value;
	double t0,t1;
	int i,k;
	
	MPI_Barrier(MPI_COMM_WORLD);

	t0 = MPI_Wtime();

	for (i = 0; i < nfiles; i++)
	{
		files[i] = fopen(infilenames[i], "r" );

		if (files[i] == NULL)
		{
			fprintf(stderr,"\nfopen() failed to open %s for reading\n", infilenames[i]);
			MPI_Abort(MPI_COMM_WORLD,1);
			exit(1);
		}

		if(use_direct_io)
			fd = fileno(files[i]);

		units_left = fsize;
		expected_value = infile_checkbyte[i];


#ifdef DEBUG
		printf("process %d started to read from file %s, file size = %d, buffer is %d\n",myid,infilenames[i],fsize,buf_size);
		fflush(stdout);
#endif

		while(units_left > 0)
		{
			units_to_do = (units_left > buf_size) ? buf_size : units_left;

			if(use_direct_io)
			{
				bytes_to_do = units_to_do * fsize_unit;
				bytes_done = read(fd,buffer,bytes_to_do);
				if(bytes_done != bytes_to_do)
				{
					fprintf(stderr,"\nERROR on %d (%s). read() = %d failed to read %d bytes from %s (fsize=%d, units_left=%d)\nERROR %d: %s\n",myid,nodename,(int)bytes_done,(int)bytes_to_do,infilenames[i],fsize,units_left,errno,strerror(errno));
					MPI_Abort(MPI_COMM_WORLD,1);
					exit(1);
				}
			}
			else
			{
				units_done = fread( buffer, fsize_unit, units_to_do, files[i]);    
				if(units_done != units_to_do)
				{
					fprintf(stderr,"\nERROR on %d (%s). fread() = %d failed to read %u units from %s (fsize=%d, units_left=%d)\nERROR %d: %s\n",myid,nodename,units_done,units_to_do,infilenames[i],fsize,units_left,errno,strerror(errno));
					MPI_Abort(MPI_COMM_WORLD,1);
					exit(1);
				}
			}
			/*     */
			if(test_correctness)
			for(k = 0; k < buffer_bytes; k ++)
			{
				if(buffer[k] != expected_value)
				{
					fprintf(stderr,"\ntest_reading() critical warning: buffer contains unexpected byte value\n");
					fprintf(stderr,"myid = %d\tfile: %s\texpected: %c (%d)\tbuffer value = %c (%d)\tat: %d\n",
						myid,infilenames[i],expected_value,expected_value,buffer[k],buffer[k],k);
					MPI_Abort(MPI_COMM_WORLD,1);
					exit(1);
				}
			}

			units_left -= units_to_do;
		}

		fclose(files[i]);
	 } 

	MPI_Barrier(MPI_COMM_WORLD);

	t1 = MPI_Wtime();

	return t1 - t0;	
}

